use std::path::PathBuf;

use based::request::RequestContext;
use maud::html;

use crate::{blacklist::check_blacklist, favicon::download_fav_for, pages::component::render_page};

/// Read directory entries into `Vec<String>`
pub fn read_dir(dir: &PathBuf) -> Vec<String> {
    let mut list = Vec::new();

    if let Ok(entries) = std::fs::read_dir(dir) {
        for entry in entries.flatten() {
            if let Some(file_name) = entry.file_name().to_str() {
                list.push(file_name.to_string());
            }
        }
    }

    list
}

/// Rewrite all URLs in `input` to the format `/s/<domain>/<path..>`
fn internalize_urls(input: &str) -> String {
    let url_pattern = r"https?://([a-zA-Z0-9.-]+)(/[\w./-]*)";
    let re = regex::Regex::new(url_pattern).unwrap();

    re.replace_all(input, |caps: &regex::Captures| {
        format!(
            "/s/{}/{}",
            &caps[1].trim_start_matches("www."), // Domain
            &caps[2]                             // Path
        )
    })
    .to_string()
}

/// Represents a directory containg archived websites
#[derive(Debug, Clone)]
pub struct WebsiteArchive {
    pub dir: PathBuf,
}

/// Represents a domain within the website archive
pub struct Domain {
    /// Domain name
    pub name: String,
    dir: PathBuf,
}

impl Domain {
    /// Creates a new `Domain` instance.
    ///
    /// If the domain name is not blacklisted, a directory is created.
    ///
    /// # Parameters
    /// - `name`: The name of the domain.
    /// - `dir`: The directory path for the domain.
    ///
    /// # Returns
    /// A new `Domain` instance.
    pub fn new(name: &str, dir: PathBuf) -> Self {
        if !check_blacklist(name) {
            std::fs::create_dir_all(&dir).unwrap();
        }
        Self {
            name: name.to_string(),
            dir,
        }
    }

    /// Resolves a specific path within the domain and returns a `Document` representing it.
    ///
    /// # Parameters
    /// - `path`: The path to resolve within the domain.
    ///
    /// # Returns
    /// A `Document` instance corresponding to the given path.
    pub fn path(&self, path: &str) -> Document {
        Document::new(&self.name, path, self.dir.parent().unwrap().to_path_buf())
    }

    /// Retrieves entries and metadata for a given path within the domain.
    ///
    /// # Parameters
    /// - `path`: The path to inspect.
    ///
    /// # Returns
    /// A tuple containing:
    /// - A vector of `PathEntry` instances representing the contents of the path.
    /// - A boolean indicating whether the path is itself a `Document`
    pub fn paths(&self, path: &str) -> (Vec<PathEntry>, bool) {
        let mut base_path = self.dir.clone();

        for p in path.split('/') {
            base_path = base_path.join(p);
        }

        let dir_content = read_dir(&base_path);

        let mut ret = Vec::new();

        let mut is_doc = false;

        for entry in dir_content {
            let url_path = format!("{path}/{entry}");
            if entry.starts_with("index_") && entry.ends_with(".html") {
                is_doc = true;
                continue;
            }

            ret.push(PathEntry(self.name.clone(), url_path));
        }

        (ret, is_doc)
    }
}

/// Represents an entry within a domain's path, containing its name and URL path.
pub struct PathEntry(String, String);

impl PathEntry {
    pub fn url(&self) -> String {
        format!("/d/{}/{}", self.0, self.1)
    }

    pub fn path(&self) -> &String {
        &self.1
    }
}

/// Represents a document within a domain
pub struct Document {
    /// The domain associated with the document.
    pub domain: String,
    /// The path of the document within the domain.
    pub path: String,
    base_dir: PathBuf,
}

impl Document {
    /// Creates a new `Document` instance.
    ///
    /// # Parameters
    /// - `domain`: The domain to which the document belongs.
    /// - `path`: The path of the document within the domain.
    /// - `base_dir`: The base directory of the archive storage.
    ///
    /// # Returns
    /// A new `Document` instance.
    pub fn new(domain: &str, path: &str, base_dir: PathBuf) -> Self {
        Self {
            domain: domain.to_string(),
            path: path
                .split('/')
                .filter(|x| !x.is_empty())
                .collect::<Vec<&str>>()
                .join("/"),
            base_dir,
        }
    }

    /// Renders the document, returning its content as a string.
    ///
    /// If the environment variable `$ROUTE_INTERNAL` is set to `true`, all links will be rewritten to point to internal archived routes.
    ///
    /// # Parameters
    /// - `version`: An optional version of the document to render in the format `YYYY-MM-DD`.
    ///
    /// # Returns
    /// An `Option` containing the rendered content as a string, or `None` if nothing could be rendered.
    pub async fn render_local(&self, version: Option<String>) -> Option<String> {
        if check_blacklist(&self.domain) {
            let content = html! {
                h3 { "This site is blacklisted" };
            };
            return Some(render_page(content, RequestContext::default()).await.1 .1);
        }

        let mut file_path = self.doc_dir();

        let latest_version = if let Some(version) = version {
            format!("index_{version}.html")
        } else {
            let versions = self.versions();
            versions.first().cloned()?
        };

        file_path = file_path.join(latest_version);

        let content = std::fs::read_to_string(file_path).ok()?;

        if std::env::var("ROUTE_INTERNAL").unwrap_or("false".to_string()) == "true" {
            Some(internalize_urls(&content))
        } else {
            Some(content)
        }
    }

    /// Determines the directory where the document is stored.
    ///
    /// # Returns
    /// A `PathBuf` representing the document directory.
    pub fn doc_dir(&self) -> PathBuf {
        let mut file_path = self.base_dir.join(&self.domain);

        for p in self.path.split('/') {
            file_path = file_path.join(p);
        }

        file_path
    }

    /// Retrieves available versions of the document.
    ///
    /// # Returns
    /// A vector of strings representing the available versions of the document, sorted in descending order.
    pub fn versions(&self) -> Vec<String> {
        let mut res: Vec<String> = read_dir(&self.doc_dir())
            .into_iter()
            .filter_map(|x| {
                if x.starts_with("index_") && x.ends_with(".html") {
                    return Some(
                        x.trim_start_matches("index_")
                            .trim_end_matches(".html")
                            .to_string(),
                    );
                }

                None
            })
            .collect();
        res.sort();
        res.reverse();
        res
    }
}

impl WebsiteArchive {
    /// Creates a new `WebsiteArchive` instance.
    ///
    /// # Parameters
    /// - `dir`: The directory path where the archive will be stored.
    ///
    /// # Returns
    /// A new `WebsiteArchive` instance.
    pub fn new(dir: &str) -> Self {
        Self {
            dir: PathBuf::from(dir),
        }
    }

    /// Retrieves the list of domain names stored in the archive.
    ///
    /// # Returns
    /// A vector of domain names as strings.
    pub fn domains(&self) -> Vec<String> {
        read_dir(&self.dir)
    }

    /// Retrieves a `Domain` instance for a specified domain name.
    ///
    /// # Parameters
    /// - `domain`: The name of the domain to retrieve.
    ///
    /// # Returns
    /// A `Domain` instance corresponding to the specified domain.
    pub fn get_domain(&self, domain: &str) -> Domain {
        Domain::new(domain, self.dir.join(domain))
    }

    /// Archives a URL by downloading and storing its content.
    ///
    /// If the URL does not pass the blacklist check, it will not be archived.
    ///
    /// # Parameters
    /// - `url`: The URL to archive.
    ///
    /// This function downloads the content of the URL, processes it, and saves it to the archive.
    pub async fn archive_url(&self, url: &str) {
        let parsed_url = url::Url::parse(url).unwrap();

        let domain = parsed_url.domain().unwrap().trim_start_matches("www.");

        // Deny blacklist
        if check_blacklist(domain) {
            return;
        }

        let path = parsed_url.path();

        let mut folder_name = self.dir.join(domain);

        download_fav_for(domain).await;

        for paths in path.split('/') {
            if !paths.is_empty() {
                folder_name = folder_name.join(paths);
            }
        }

        std::fs::create_dir_all(&folder_name).unwrap();

        let timestamp = chrono::Utc::now().format("%Y-%m-%d").to_string();
        let filename = folder_name.join(format!("index_{timestamp}.html"));

        log::info!("Archiving {url} to {}", filename.to_str().unwrap());

        run_command(&[
            "monolith",
            "-I",
            "-o",
            filename.to_str().unwrap(),
            &format!("https://{}/{}", domain, path),
        ]);
    }
}

// full text search
// add new sites?
// transparent auto page downloading
// redownload after threshold

fn run_command(cmd: &[&str]) {
    let mut cmd_setup = std::process::Command::new(cmd[0]);
    let cmd_setup = cmd_setup
        .args(cmd.iter().skip(1).collect::<Vec<_>>())
        .stdout(std::process::Stdio::inherit())
        .stderr(std::process::Stdio::inherit());

    let child = cmd_setup.spawn().unwrap();

    let status = child.wait_with_output().unwrap();
    assert!(status.status.success());
}
